//
//  WalletConnection.swift
//
//
//  Created by Pat Nakajima on 11/22/22.
//

import Foundation
import UIKit
import WalletConnectSwift
import web3

extension WCURL {
	var asURL: URL {
		// swiftlint:disable force_unwrapping
		URL(string: "wc://wc?uri=\(absoluteString)")!
		// swiftlint:enable force_unwrapping
	}
}

enum WalletConnectionError: String, Error {
	case walletConnectURL
	case noSession
	case noAddress
	case invalidMessage
	case noSignature
}

protocol WalletConnection {
	var isConnected: Bool { get }
	var walletAddress: String? { get }
	func preferredConnectionMethod() throws -> WalletConnectionMethodType
	func connect() async throws
	func sign(_ data: Data) async throws -> Data
}

class WCWalletConnection: WalletConnection, WalletConnectSwift.ClientDelegate {
	@Published var isConnected = false

	var walletConnectClient: WalletConnectSwift.Client!
	var session: WalletConnectSwift.Session? {
		didSet {
			DispatchQueue.main.async {
				self.isConnected = self.session != nil
			}
		}
	}

	init() {
		let peerMeta = Session.ClientMeta(
			name: "xmtp-ios",
			description: "XMTP",
			icons: [],
			// swiftlint:disable force_unwrapping
			url: URL(string: "https://safe.gnosis.io")!
			// swiftlint:enable force_unwrapping
		)
		let dAppInfo = WalletConnectSwift.Session.DAppInfo(peerId: UUID().uuidString, peerMeta: peerMeta)

		walletConnectClient = WalletConnectSwift.Client(delegate: self, dAppInfo: dAppInfo)
	}

	func preferredConnectionMethod() throws -> WalletConnectionMethodType {
		guard let url = walletConnectURL else {
			throw WalletConnectionError.walletConnectURL
		}

		if let url = URL(string: url.absoluteString), UIApplication.shared.canOpenURL(url) {
			return WalletRedirectConnectionMethod(redirectURI: url.absoluteString).type
		}

		return WalletQRCodeConnectionMethod(redirectURI: url.absoluteString).type
	}

	lazy var walletConnectURL: WCURL? = {
		do {
			let keybytes = try Crypto.secureRandomBytes(count: 32)

			return WCURL(
				topic: UUID().uuidString,
				// swiftlint:disable force_unwrapping
				bridgeURL: URL(string: "https://bridge.walletconnect.org")!,
				// swiftlint:enable force_unwrapping
				key: keybytes.toHex
			)
		} catch {
			return nil
		}
	}()

	func connect() async throws {
		guard let url = walletConnectURL else {
			throw WalletConnectionError.walletConnectURL
		}

		try walletConnectClient.connect(to: url)
	}

	func sign(_ data: Data) async throws -> Data {
		guard session != nil else {
			throw WalletConnectionError.noSession
		}

		guard let walletAddress = walletAddress else {
			throw WalletConnectionError.noAddress
		}

		guard let url = walletConnectURL else {
			throw WalletConnectionError.walletConnectURL
		}

		guard let message = String(data: data, encoding: .utf8) else {
			throw WalletConnectionError.invalidMessage
		}

		return try await withCheckedThrowingContinuation { continuation in
			do {
				try walletConnectClient.personal_sign(url: url, message: message, account: walletAddress) { response in
					if let error = response.error {
						continuation.resume(throwing: error)
						return
					}

					do {
						let resultString = try response.result(as: String.self)

						guard let resultDataBytes = resultString.web3.bytesFromHex else {
							continuation.resume(throwing: WalletConnectionError.noSignature)
							return
						}

						let resultData = Data(resultDataBytes)
						continuation.resume(returning: resultData)
					} catch {
						continuation.resume(throwing: error)
					}
				}
			} catch {
				continuation.resume(throwing: error)
			}
		}
	}

	var walletAddress: String? {
		if let address = session?.walletInfo?.accounts.first {
			return EthereumAddress(address).toChecksumAddress()
		}

		return nil
	}

	func client(_: WalletConnectSwift.Client, didConnect _: WalletConnectSwift.WCURL) {}

	func client(_: WalletConnectSwift.Client, didFailToConnect _: WalletConnectSwift.WCURL) {}

	func client(_: WalletConnectSwift.Client, didConnect session: WalletConnectSwift.Session) {
		// TODO: Cache session
		self.session = session
	}

	func client(_: WalletConnectSwift.Client, didUpdate session: WalletConnectSwift.Session) {
		self.session = session
	}

	func client(_: WalletConnectSwift.Client, didDisconnect _: WalletConnectSwift.Session) {
		session = nil
	}
}
